
//*******************************************************************
//
//  UNIT:     B l u r                       06.09.99,  17.04.2000
//
//  bluring routines for delphi-4 bitmaps
//  coded by Gerd Platl                                            
//
//-------------------------------------------------------------------
//  Blur2H              bluring with 2 horizontal pixels
//  Blur2L              bluring with 2 horizontal pixels
//  Blur2M              bluring with 2 horizontal pixels
//  Blur3H              bluring with 3 horizontal pixels
//  Blur3S              bluring with 3 horizontal pixels with direction
//  Blur4Q              bluring with 2x2 pixels
//  Blur4C              bluring with cross pixels
//  Blur4L              bluring with cross pixels one line up
//
//*******************************************************************
unit Blur;

interface

uses Graphics;

procedure Blur2H (source, dest: tBitmap; y1,y2: cardinal);
procedure Blur2L (source, dest: tBitmap; y1,y2: cardinal);
procedure Blur2M (source, dest: tBitmap; y1,y2: cardinal);
procedure Blur3H (source, dest: tBitmap; y1,y2: cardinal);
procedure Blur3S (source, dest: tBitmap; y1,y2,d: cardinal);
procedure Blur4Q (source, dest: tBitmap; y1,y2: cardinal);
procedure Blur4C (source, dest: tBitmap; y1,y2: cardinal);
procedure Blur4L (source, dest: tBitmap; y1,y2: cardinal);

//*********************************************************
implementation

//*********************************************************
// input parameters of all bluring routines:
//   source: tBitmap      source bitmap
//   dest: tBitmap        destination bitmap
//   y1,y2: cardinal      first and last line index
//   d: cardinal          vertical destination delta
//---------------------------------------------------------

//*********************************************************
// blures left pixel                x  o
// of 2 pixels                     p1  sp^
//---------------------------------------------------------
procedure Blur2H (source, dest: tBitmap; y1,y2: cardinal);
var  xi, yi: cardinal;
     p1: byte;
     sp, dp: ^byte;
begin
  for yi := y1 to y2
  do begin
    sp := source.ScanLine[yi];
    dp :=   dest.ScanLine[yi];
    for xi := 1 to source.Width
    do begin
      p1 := sp^;
      inc (sp);
      dp^ := (p1 + sp^) shr 1;
      inc (dp);
    end;
  end;
end;
//*********************************************************
// blures left pixel                x  o
// of 2 pixels with decrement      p1  sp^
//---------------------------------------------------------
procedure Blur2L (source, dest: tBitmap; y1,y2: cardinal);
var  xi, yi: cardinal;
     p1, p2: byte;
     sp, dp: ^byte;
begin
  for yi := y1 to y2
  do begin
    sp := source.ScanLine[yi];
    dp :=   dest.ScanLine[yi];
    for xi := 1 to source.Width
    do begin
      p1 := sp^;
      inc (sp);
      p2 := (p1 + sp^) shr 1;
      if p2 > 0 then dp^ := pred(p2)
      else dp^ := p2;
      inc (dp);
    end;
  end;
end;
//*********************************************************
// blures middle pixel              o  x  o
// of 2 pixels                     p1 p2 sp^
//---------------------------------------------------------
procedure Blur2M (source, dest: tBitmap; y1,y2: cardinal);
var  xi, yi: cardinal;
     p1, p2, p3: byte;
     sp, dp: ^byte;
begin
  for yi := y1+1 to y2
  do begin
    sp := source.ScanLine[yi];
    dp :=   dest.ScanLine[yi];  inc (dp);
    p1 := sp^;   inc (sp);
    p2 := sp^;   inc (sp);
    for xi := 2 to source.Width
    do begin
      p3 := sp^;
      dp^ := (p1 + sp^) shr 1;
      inc (sp);
      inc (dp);
      p1 := p2;
      p2 := p3;
    end;
  end;
end;
//*********************************************************
// blures middle pixel              o  x  o
// of 3 pixels with direction       p1 p2 sp^
//---------------------------------------------------------
procedure Blur3S (source, dest: tBitmap; y1,y2, d: cardinal);
var  xi, yi: cardinal;
     p1, p2: byte;
     sp, dp: ^byte;
begin
  if y1 < d then y1 := d;
  for yi := y1 to y2
  do begin
    sp := source.ScanLine[yi];
    dp :=   dest.ScanLine[yi-d];
    inc (dp);
    p1 := sp^;
    inc (sp);
    for xi := 2 to source.Width
    do begin
      p2 := sp^;
      inc (sp);
      dp^ := (p1 + p2 shl 1 + sp^) shr 2;
//    dp^ := (p1 + p2 shl 1 + sp^ + 1) shr 2;
      inc (dp);
      p1 := p2;
    end;
  end;
end;
//*********************************************************
// blures middle pixel              o  x  o
// of 3 pixels                     p1 p2 sp^
//---------------------------------------------------------
procedure Blur3H (source, dest: tBitmap; y1,y2: cardinal);
begin
  Blur3S (source, dest, y1,y2, 0);
end;
//*********************************************************
// blures left upper pixel          s1:   x  o
// of 2x2 pixel square              s2:   o  o
//                                        p1 p2
//---------------------------------------------------------
procedure Blur4Q (source, dest: tBitmap; y1,y2: cardinal);
var  xi, yi: cardinal;
     p1, p2: word;
     s1, s2, dp: ^byte;
begin
  s1 := source.ScanLine[y1];
  for yi := y1 to y2-1
  do begin
    s2 := source.ScanLine[yi+1];
    dp :=   dest.ScanLine[yi];
    p1 := s1^ + s2^;
    inc (s1);
    inc (s2);
    for xi := 2 to source.Width
    do begin
      p2 := s1^ + s2^;
      inc (s1);
      inc (s2);
      dp^ := (p1 + p2) shr 2;
      inc (dp);
      p1 := p2;
    end;
    s1 := s2;
  end;
end;
//*********************************************************
// blures middle pixel              s1:      o
// of cross figure                  s2:   o  x  o
//                                  s3:      o
//                                        p1 p2 p3
//---------------------------------------------------------
procedure Blur4C (source, dest: tBitmap; y1,y2: cardinal);
var  xi, yi: cardinal;
     p1,p2,p3: byte;
     l1,l2,l3, s1,s2,s3, dp: ^byte;
begin
  l1 := source.ScanLine[y1];
  l2 := source.ScanLine[y1+1];
  for yi := y1 to y2
  do begin
    if yi = y2
    then l3 := source.ScanLine[yi]
    else l3 := source.ScanLine[yi+1];
    s1 := l1;
    s2 := l2;
    s3 := l3;
    dp := dest.ScanLine[yi];
    p1 := s2^;   inc (s2);
    p2 := s2^;
    for xi := 2 to source.Width
    do begin
      inc (s1);
      inc (s2);
      inc (s3);
      inc (dp);
      p3 := s2^;
      dp^ := (s1^ + p1 + p3 + s3^) shr 2;
      p1 := p2;
      p2 := p3;
    end;
    l1 := l2;
    l2 := l3;
  end;
end;
//*********************************************************
// blures left pixel                s1:      o
// of cross figure                  s2:   x  o  o
//                                  s3:      o
//                                        p1 p2 p3
//---------------------------------------------------------
procedure Blur4L (source, dest: tBitmap; y1,y2: cardinal);
var  xi, yi: cardinal;
     p1,p2,p3: byte;
     l1,l2,l3, s1,s2,s3, dp: ^byte;
begin
  l1 := source.ScanLine[y1];
  l2 := source.ScanLine[y1];
  for yi := y1 to y2
  do begin
    if yi = y2
    then l3 := source.ScanLine[yi]
    else l3 := source.ScanLine[yi+1];
    s1 := l1;
    s2 := l2;
    s3 := l3;
    dp := dest.ScanLine[yi];
    p1 := s2^;
    inc (s2);
    for xi := 2 to source.Width
    do begin
      p2 := s2^;
      p3 := (p1 + s1^ + s2^ + s3^) shr 2;
      if p3 > 0
      then dp^ := pred(p3)
      else dp^ := 0;
      inc (s1);
      inc (s2);
      inc (s3);
      inc (dp);
      p1 := p2;
    end;
    p3 := (s1^ + s3^) shr 1;
    if p3 > 0
    then dp^ := pred(p3)
    else dp^ := 0;
    l1 := l2;
    l2 := l3;
  end;
end;

end.

