
//*********************************************************
//
//  UNIT:  F r a c t M a p             14.09.98, 05.06.2003
//
//  create fractal map with dimension 2^nDim
//  coded by Gerd Platl
//
//  the fractal map consist of a height and a color map
//
//---------------------------------------------------------
//  CreateMap           create new height and color map
//  DeleteMap           delete height and color map
//  SelectMap           select height and color map
//  SaveMap             save height and color map files
//  LoadMap             load height and color map files
//
//ToDo: map generation with dynamic dimensions
//
//*********************************************************
{$I 'comp_opt.inc'}

unit FractMap;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
  {$define xtest3    help functions}
{$endif}

interface

uses  Windows, Graphics, Graph, Palette;

const heightMapFileExt = 'H.BMP';  // file extension
      colorMapFileExt  = 'C.BMP';

const xFac   = 8;            // size value: 8-> Dim=256, maxX=255
      yFac   = 8;
      xShift = 16 - xFac;    // right shift value
      yShift = 16 - yFac;
      xDim   = 1 shl xFac;   // size: must be 2^Fac
      yDim   = 1 shl yFac;
      maxX   = xDim - 1;     // max values
      maxY   = yDim - 1;

type  tColorArray = array[0..maxColor] of tColor;

var   heightMap: tBitmap = nil;
      colorMap:  tBitmap = nil;
      viewMap:   tBitmap = nil;

procedure CreateMap (var hm, cm: tBitmap);
procedure DeleteMap (var hm, cm: tBitmap);
procedure SelectMap (var hm, cm: tBitmap);
procedure SaveMap (filename: string);
procedure LoadMap (filename: string);
procedure GetViewMapPosition (var xi,yi, xv,yv: word);

function  GetHeight (x,y: word): byte;
function  GetColor  (x,y: word): tColor;
procedure PutHeight (x,y: integer; height: byte);
procedure PutColor  (x,y: integer; color: tColor);
procedure ChangeColorMap;
procedure CorrectHeightMap;
procedure GenerateMap (fractalDim: integer);

procedure GenerateFractal (no: integer);

//*********************************************************
implementation

uses  SysUtils, Classes, Images,
      SysUtil2;

//=========================================================
// palette handling
//=========================================================

const spreadColors = 12;
      minH = 85;   // minimum height = sea level

const fmColors: array [0..spreadColors-1] of tColorRGBI
      =((rb:  0; gb:  0; bb:120; ii:   0)   {dunkelblau}
       ,(rb: 70; gb: 90; bb:255; ii:  50)   {blau}
       ,(rb:150; gb:150; bb:233; ii:  80)   {hellblau}
       ,(rb:235; gb:225; bb: 80; ii:minH)   {hellbeige}
       ,(rb:200; gb:160; bb: 70; ii: 100)   {dunkelbeige}
       ,(rb: 50; gb:255; bb:  0; ii: 120)   {hellgrn}
       ,(rb: 55; gb:133; bb: 55; ii: 140)   {dunkelgrn}
       ,(rb:111; gb:111; bb: 33; ii: 180)   {braun}
       ,(rb: 77; gb: 77; bb: 11; ii: 185)   {dunkelgrau}
       ,(rb:160; gb:160; bb:160; ii: 210)   {grau}
       ,(rb:215; gb:215; bb:215; ii: 235)   {hellgrau}
       ,(rb:255; gb:255; bb:255; ii: 255)); {weiss}

var   sPal: tPalette;        // standard color palette
      hPal: tPalette;        // height map palette
      cPal: tPalette;        // color map palette
      cMap: tColorArray;     // palette with tColor

//*********************************************************
// create color palette for color map
//---------------------------------------------------------
procedure InitLandscapeColors (var pal: tPalette);
begin
{$ifdef test1}
  writeln (tst, 'FractMap.InitLandscapeColors: ');
{$endif}

  SetSpreadPalette (pal, fmColors, spreadColors);
  if sizeof (pal) = sizeof(cMap)
  then Move (pal, cMap, sizeof (cMap));
end;
//*********************************************************
// create view map and set color palette
//---------------------------------------------------------
procedure InitViewMap (x,y: word);
begin
{$ifdef test1}
  writeln (tst, 'FractMap.InitViewMap: ');
{$endif}

  viewMap := tBitmap.Create;
  viewMap.PixelFormat := pf8Bit;
  viewMap.Width  := x;
  viewMap.Height := y;
  SetBitmapPalette (viewMap, cPal);
end;

//=========================================================
// fractal map handling
//=========================================================

//*********************************************************
// create new fractal height and color map
// input:      hm: tBitmap    height 8 bit map
//             cm: tBitmap    color 8 bit map (256 colors)
//---------------------------------------------------------
procedure CreateMap (var hm, cm: tBitmap);
begin
{$ifdef test1}
  writeln (tst, 'FractMap.CreateMap (',HexString(hm),', ', HexString(cm),')');
{$endif}

  // height map
  if hm = nil
  then begin
    hm := tBitmap.Create;
    hm.PixelFormat := pf8Bit;
    hm.Width  := xDim;
    hm.Height := yDim;
    SetBitmapPalette (hm, hPal);
  end;

  // color map
  if cm = nil
  then begin
    cm := tBitmap.Create;
    cm.PixelFormat := pf8Bit;
    cm.Width  := xDim;
    cm.Height := yDim;
    SetBitmapPalette (cm, cPal);
  end;

  SelectMap (hm, cm);
end;
//---------------------------------------------------------
procedure DeleteMap (var hm, cm: tBitmap);
begin
{$ifdef test1}
  writeln (tst, 'FractMap.DeleteMap (',HexString(hm),', ', HexString(cm),')');
{$endif}

  FreeAndNil (hm);   // hm.Free;   hm := nil;
  FreeAndNil (cm);   // cm.Free;   cm := nil;
end;
//*********************************************************
// copy bitmap pointers to global variables
// and set color map palette as active palette
//---------------------------------------------------------
procedure SelectMap (var hm, cm: tBitmap);
begin
{$ifdef test2}
  writeln (tst, 'FractMap.SelectMap (',HexString(hm),', ', HexString(cm),')');
{$endif}
  heightMap := hm;
  colorMap  := cm;
  ClearBitmap (bgBitmap);
  // set palette of color map to background bitmap
  SetBitmapPalette (bgBitmap, cPal);
  SetAllPalette (cPal);
end;
//*********************************************************
// create view map from active color map
//---------------------------------------------------------
procedure PrepareViewMap (vPal: tPalette);
begin
  ClearBitmap (viewMap);
  SetBitmapPalette (viewMap, vPal);
  StretchBitMap (colorMap, viewMap);
end;
//*********************************************************
//  save height and color map files
//---------------------------------------------------------
procedure SaveMap (filename: string);
begin
  SaveBMPBitmap (heightMap, filename+heightMapFileExt);
  SaveBMPBitmap (colorMap,  filename+colorMapFileExt);
end;
//*********************************************************
//  load height and color map files and create view map
//---------------------------------------------------------
procedure LoadMap (filename: string);
var  hm,cm: tBitmap;
begin
{$ifdef test1}
  writeln (tst, 'FractMap.LoadMap: ', filename);
{$endif}

  LoadBitmap (hm, filename + heightMapFileExt);
  LoadBitmap (cm, filename + colorMapFileExt);
  GetBitmapPalette (cm, cPal);
  ClearBitmap (colorMap);
  SetBitmapPalette (colorMap, cPal);
  StretchBitMap (cm, colorMap);
  StretchBitMap (hm, heightMap);
  PrepareViewMap (cPal);
  SetAllPalette (cPal);
//  writeln (tst, '   mapDim: ', colorMap.width, '*', colorMap.height);
end;

//*********************************************************
// transform integer position to view map position
// input:    xi,yi    integer position 0..maxWord (65535)
// output:   xi,yi    view map position
//---------------------------------------------------------
procedure GetViewMapPosition (var xi,yi, xv,yv: word);
begin
  xv := (longint(xi) * viewMap.Width)  div maxWord;
  yv := (longint(yi) * viewMap.Height) div maxWord;
end;
//=========================================================
// additional routines
//=========================================================

//---------------------------------------------------------
function GetHeight (x,y: word): byte;
var   linePtr: pByteArray;
begin
  linePtr := heightMap.ScanLine[y and maxY];
  GetHeight := linePtr^[x and maxX];
end;
//---------------------------------------------------------
function GetColor (x,y: word): tColor;
var      linePtr: pByteArray;
begin
  linePtr := heightMap.ScanLine[y and maxY];
  GetColor := cMap[linePtr[x and maxX]];
end;
//---------------------------------------------------------
procedure PutHeight (x,y: integer; height: byte);
var   linePtr: pByteArray;
begin
  linePtr := heightMap.ScanLine[y and maxY];
  linePtr^[x and maxX] := height;

//  sleep(1);   {only for testing}
//  activeCanvas.Pixels[x,y] := height;
end;
//---------------------------------------------------------
procedure PutColor (x,y: integer; color: tColor);
begin
  // not implemented yet
end;

//*********************************************************
// ToDo: implement ChangeColorMap
//---------------------------------------------------------
procedure ChangeColorMap;
var   xi,yi: word;  rcol: tColor;
begin
  for yi := 0 to maxY
  do begin
    for xi := 0 to maxX
    do begin
      rcol := longint(cMap[xi and $FF]) - 8 + GetColor (xi+44,yi+55) div 16;
      if      rcol <  16 then rcol := 16
      else if rcol > 255 then rcol := 255;
      cMap[xi and $FF] := rcol;
    end;
  end;
end;
//---------------------------------------------------------
procedure CopyHeight2ColorMap;
begin
  CopyBitmap (heightMap, colorMap);
end;
//*********************************************************
// set minimum height to sea level
//---------------------------------------------------------
procedure CorrectHeightMap;
var   xi,yi: word;  lp: pByteArray;
begin
  for yi := 0 to maxY
  do begin
    lp := heightMap.Scanline[yi];
    for xi := 0 to maxX
    do if lp^[xi] < minH then lp^[xi] := minH;
  end;
end;
//*********************************************************
// create new random map
// input:    fractalDim:   fractal dimension 1..100 %
//---------------------------------------------------------
procedure RandomMap (fractalDim: integer);

  procedure RandomHeight(x1,y1,x2,y2,max:word);
  var col: integer;
  begin
    col := (integer (GetHeight(x1,y1))
           +integer (GetHeight(x2,y2)))shr 1;
    col := col + random(max) - random(max);
    if col > 250 then col := 250
    else if col < 16 then col := 16;
    x1 := (x1+x2) shr 1;
    y1 := (y1+y2) shr 1;
    if (x1 <= maxX) and (y1 <= maxY)
    then PutHeight (x1,y1,col);
    if x1 = 0 then PutHeight (xDim,y1,col);
    if y1 = 0 then PutHeight (x1,yDim,col);
  end;

  procedure GenSub(x1,y1,x3,y3,max:word);
  var x2,y2:word;
  begin
    x2 := (x1+x3)shr 1;
    y2 := (y1+y3)shr 1;
    max := integer(max) * fractalDim div 100;
    if x1 = 0 then RandomHeight (0,y1,0,y3,max); {links}
    if y1 = 0 then RandomHeight (x1,0,x3,0,max); {oben}
    RandomHeight (x1,y3,x3,y3,max); {unten}
    RandomHeight (x3,y1,x3,y3,max); {rechts}
    RandomHeight (x2,y1,x2,y3,max); {mitte}
    if x2-x1 > 1
    then begin
      GenSub (x1,y1,x2,y2,max);
      GenSub (x2,y1,x3,y2,max);
      GenSub (x1,y2,x2,y3,max);
      GenSub (x2,y2,x3,y3,max);
    end;
  end;

begin
{$ifdef test3}
  writeln (tst, 'FractMap.RandomMap (', fractalDim, ')');
{$endif}

  PutHeight (   0,    0, 128+random(128));
  PutHeight (xDim, yDim,  GetHeight(0,0));
  PutHeight (   0, yDim,  GetHeight(0,0));
  PutHeight (xDim,    0,  GetHeight(0,0));
  GenSub (0,0, xDim,yDim, 255 * 100 div fractalDim);
end;

//*********************************************************
type  float = double; {extended;}
      tValues = record
        Xmin, Xmax: float;
        Ymin, Ymax: float;
        Xc,   Yc  : float;
        Xmid, Ymid: float;
      end;

const maxIt: word = 255;               {interations}

var   xpos,ypos  : float;     {akt. Float-Position}

const mv1: tValues
      =(Xmin: -2.25;
        Xmax: +1.0;
        Ymin: -1.62;
        Ymax: +1.62;
        Xc  :  0.5;
        Yc  :  0.5);

const mv2: tValues
      =(Xmin: -1.1935596;
        Xmax: -1.1935433;
        Ymin: +0.3069270;
        Ymax: +0.3069378;
        Xc  :  0.5;
        Yc  :  0.5);

const mv3: tValues
      =(Xmin: 0.279291522;
        Xmax: 0.279291605;
        Ymin: 0.605386100;
        Ymax: 0.605386038;
        Xc  :  0.5;
        Yc  :  0.5);

//---------------------------------------------------------
function CalcMandel1: byte;

var  iter     :integer;
     A,B,a2,b2:float;
begin
  A:=0;
  B:=0;
//  for iter := 0 to maxIt
  for iter := maxIt downto 0
  do begin
    a2:=sqr(A);
    b2:=sqr(B);
    B:= 2*A*B + ypos;
    A:= a2-b2 + xpos;
    if (a2+b2) > maxIt then break
  end;
  CalcMandel1 := byte(iter);
END;
//---------------------------------------------------------
procedure Mandel1 (var mv: tValues; Xa,Xe, Ya,Ye: integer);
var
     xp,yp: integer;   {akt. Pixel-Position}
     dx,dy: float;     {Schrittweite}
begin
  with mv
  do begin
    dx := (Xmax - Xmin) / (xe - xa + 1);
    dy := (Ymax - Ymin) / (ye - ya + 1);
    yp := ya;
    ypos := ymin;
    repeat
      xpos := xmin;
      for xp := xa to xe
      do begin
        PutHeight (xp,yp, CalcMandel1);
        xpos := xpos + dx;
      end;
      inc (yp);
      ypos := ypos + dy;
    until yp > ye;
  end;
  BitmapMirror4 (heightMap);
end;
//*********************************************************
// generate new fractal and view map
// input:    fractalDim:   fractal dimension 1..100 %
//---------------------------------------------------------
procedure GenerateMap (fractalDim: integer);
begin
{$ifdef test1}
  writeln (tst, 'FractMap.GenerateMap (', fractalDim, ')');
{$endif}
  RandomMap (fractalDim);
  CopyHeight2ColorMap;
  ChangeColorMap;
  CorrectHeightMap;
  cPal := sPal;
  // Absturz ???
  SetBitmapPalette (bgBitmap, sPal);
  SetAllPalette (sPal);
  PrepareViewMap (sPal);
end;
//---------------------------------------------------------
procedure GenerateFractal (no: integer);
begin
  case no of
    1: Mandel1 (mv1, 0,maxX div 2, 0,maxY div 2);
    2: Mandel1 (mv2, 0,maxX div 2, 0,maxY div 2);
    3: Mandel1 (mv3, 0,maxX div 2, 0,maxY div 2);
  end;
//  SetBitmapPalette (heightMap, hPal);
  CopyHeight2ColorMap;
//  SetBitmapPalette (colorMap, cPal);
  SetAllPalette (cPal);
  ChangeColorMap;
  CorrectHeightMap;
  PrepareViewMap (cPal);
end;

//*********************************************************
initialization
  SetIntensityPalette (hPal);  // create height map palette
  InitLandscapeColors (sPal);  // create standard color palette
  cPal := sPal;                // create color map palette
  InitViewMap (128,128);       // create view map
end.

