
//*******************************************************************
//
//  UNIT:     P a l e t t e                 22.07.99,  26.03.2003
//
//  delphi palette handling
//
//-------------------------------------------------------------------
//  InitPalette         initialize bitmap palette
//  SetAllPalette       set 256 color palette
//  GetAllPalette       return 256 color palette
//
//  SetPaletteRGB       set RGB color value of color palette
//  SetPaletteColor     set one color value of color palette
//  GetPaletteColor     return color value of given palette index
//
//  SetIntensityPalette create gray scaled palette
//  SetSineWavePalette  create sine wave color palette
//
//  DrawPalette         draw color bar of given palette
//  DrawRGBPalette      draw color bar of active palette
//
//  SetBitmapPalette    set color palette of given bitmap
//  GetBitmapPalette    get color palette of given bitmap
//  CopyBitmapPalette   copy bitmap palette to other bitmap
//  UseBitmapPalette    use palette of given bitmap
//
//  LoadPalette         load palette from file
//  SavePalette         save palette to file
//
//*******************************************************************
{$I comp_opt.inc}

unit Palette;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
{$endif}

interface

uses  Windows, SysUtils, Graphics,
      SysUtil2, Math2, ColorRGB, Graph;

type  tColorRGBI = ColorRGB.tColorRGBI;

type  tColorArray = array[0..9999] of tColorRGBI;
      pColorArray = ^tColorArray;

type  tPalette = array [0..255] of tColorRGBI;
      pPalette = ^tPalette;

(*
type  pRGBPalette = ^tRGBPalette;
      tRGBPalette = array [0..maxColorRGB] of tRGB;
*)

const cBlack  :tColorRGBI = (rb:  0; gb:  0; bb:   0);
      cRed    :tColorRGBI = (rb:255; gb:  0; bb:   0);
      cGreen  :tColorRGBI = (rb:  0; gb:255; bb:   0);
      cBlue   :tColorRGBI = (rb:  0; gb:  0; bb: 255);
      cYellow :tColorRGBI = (rb:255; gb:255; bb:   0);
      cWhite  :tColorRGBI = (rb:255; gb:255; bb: 255);

procedure InitPalette;
procedure SetAllPalette    (var pal: tPalette);
procedure GetAllPalette    (var pal: tPalette);
procedure SetSpreadPalette (var pal: tPalette;
                            var spreadValues: array of tColorRGBI; len: byte);

procedure SpreadColor         (var pal: tPalette; lastIndex: byte);
procedure SetIntensityPalette (var pal: tPalette);
procedure SetLayerPalette     (var pal: tPalette; color: tColor; layers: word);
procedure SetSineWavePalette  (var pal: tPalette; start, rs,gs,bs: word);

procedure SetPaletteRGB   (colorIndex: byte; r,g,b: byte);
procedure SetPaletteColor (colorIndex: byte; color: tColor);
 function GetPaletteColor (colorIndex: byte): tColor;

procedure DrawPalette (var pal: tPalette; x,y,h: integer);
procedure DrawRGBPalette (x,y,h: integer);

{--- bitmap handling ---}
procedure SetBitmapPalette (bm: tBitmap; var pal: tPalette);
procedure GetBitmapPalette (bm: tBitmap; var pal: tPalette);
procedure CopyBitmapPalette(bm, dm: tBitmap);
procedure UseBitmapPalette (bm: tBitmap);

{--- file handling ---}
procedure LoadPalette(var pal: tPalette; filename: String);
procedure SavePalette(var pal: tPalette; filename: String);

{I Standard.PAL    standard palette with 256 colors}
{I CCube256.PAL    color values of 256-RGB-color-RGB-cube}

//*********************************************************
implementation

const palColors = 256;
      palSize = palColors * sizeof (tColorRGBI);

var   firstSpreadIndex: byte = 0;
      activePalette: tMaxLogPalette;
      paletteHandle: HPALETTE;

//*********************************************************
// Interpolation der Farben von Farbindex first bis last
// in der bergebenen Farbpalette.
// Versorgung:  pal: tPalette;      Farbpalette
//              first, last: byte;  1..letzter Farbindex
// Ergebnis:    pal: tPalette;      neue Farbpalette
//---------------------------------------------------------
procedure SpreadColor (var pal: tPalette; lastIndex: byte);

var   ci,dc: word;   cp1, cp2: tColorRGBI;
      rd,gd,bd, rv,gv,bv: longint;
begin
  dc := lastIndex - firstSpreadIndex + 1;
  cp1 := pal[firstSpreadIndex];
  cp2 := pal[lastIndex];
  rd := cp2.rb - cp1.rb;   rv := 0;
  gd := cp2.gb - cp1.gb;   gv := 0;
  bd := cp2.bb - cp1.bb;   bv := 0;
  for ci := firstSpreadIndex + 1 to lastIndex
  do with pal[ci]
  do begin
    inc (rv,rd);   rb := cp1.rb + rv div dc;
    inc (gv,gd);   gb := cp1.gb + gv div dc;
    inc (bv,bd);   bb := cp1.bb + bv div dc;
  end;
  firstSpreadIndex := lastIndex;
end;
//*********************************************************
// create new color palette given by spread color array
// input:      pal             palette
//             spreadValues:   spread colors
//             len:            number of spread colors
//---------------------------------------------------------
procedure SetSpreadPalette (var pal: tPalette;
                            var spreadValues: array of tColorRGBI; len: byte);
var   si,ci: word;
begin
  firstSpreadIndex := spreadValues[0].ii;
  pal[firstSpreadIndex] := spreadValues[0];
  pal[firstSpreadIndex].ii := 0;
  for si := 1 to len - 1
  do begin
    ci := spreadValues[si].ii;
    pal[ci] := spreadValues[si];
    pal[ci].ii := 0;
    SpreadColor (pal, ci);
  end;
end;

{$Q-}
//*********************************************************
procedure ActivatePalette;
begin
  paletteHandle := Windows.CreatePalette(pLogPalette(@activePalette)^);
  if paletteHandle <> 0
  then bgBitmap.Palette := paletteHandle;
end;
//*********************************************************
// set gray scaled palette
// output:     pal:       gray scaled palette
//---------------------------------------------------------
procedure SetIntensityPalette (var pal: tPalette);
var   ci: Integer;
begin
  for ci := 0 to 255
  do with pal[ci]
  do begin
    rb := ci;
    gb := ci;
    bb := ci;
  end;
end;
//*********************************************************
// set colored layer palette
// output:     pal:       calculated palette
// input:      color:     main color for scaling
//             layers:    numbers of different colors
//---------------------------------------------------------
procedure SetLayerPalette (var pal: tPalette; color: tColor; layers: word);
var   ci, d,f: Integer;
begin
  d := 255 div layers;
  for ci := 0 to 255
  do with pal[ci]
  do begin
    f := (ci div d) * d;
    rb := (tColorRGBI(color).rb * f) div 255;
    gb := (tColorRGBI(color).gb * f) div 255;
    bb := (tColorRGBI(color).bb * f) div 255;
  end;
end;
//*********************************************************
// create palette with sine wave colors
// input:      start: random start range
//             rs,gs,bs:  rgb steps
// output:     pal:       sine wave palette
//---------------------------------------------------------
procedure SetSineWavePalette (var pal: tPalette; start, rs,gs,bs: word);
var   ci, rVal, gVal, bVal: word;
begin
{$ifdef test1}
  writeln (tst, 'Palette.SetSineWavePalette: ',rs:5,gs:5,bs:5);
{$endif}
  rVal := word(193*256 + random(start));
  gVal := word(193*256 + random(start));
  bVal := word(193*256 + random(start));
  for ci := 0 to 255
  do begin
    with pal[ci]
    do begin
      rb := sinB [rVal div 256];
      gb := sinB [gVal div 256];
      bb := sinB [bVal div 256];
    end;
    inc (rVal, rs);
    inc (gVal, gs);
    inc (bVal, bs);
  end;
end;

//*********************************************************
// call:       DrawPalette (pal, x,y,h);
// purpose:    draw given color palette at position x,y
// input:      pal: tPalette
//             x,y: integer;    left top position
//             h: integer;      height of color bar
//---------------------------------------------------------
procedure DrawPalette (var pal: tPalette;
                       x,y,h: integer);
var   w,ci,yi,y2: byte;
begin
  if y+h > GetMaxY then h := GetMaxY - y;
  y2 := y + h;
  SetColor (clBlack);
  Rectangle (x, y, x+257, y2+1);
  inc (x);
  inc (y);
  w := 255;
  if x+w >= pageX then w := pageX - x;

(*
  // common routine  (not running ???)
  for ci := 0 to w-1
  do begin
    SetColor (ci);
    DrawVLine (x,y, y2);
    inc (x);
  end;
*)
  // draw one palette color line
  for ci := 0 to w-1
  do PutPixel (x+ci, y, ci);

  // copy color lines
  for yi := y+1 to y2-1
  do CopyLine (bgBitmap.Canvas, bgBitmap.Canvas, x,y, x,yi, w);
end;
//*********************************************************
// call:       DrawRGBPalette (x,y,h);
// purpose:    draw active color palette at position x,y
// input:      x,y: integer;    left top position
//             h: integer;      height of color bar
//---------------------------------------------------------
procedure DrawRGBPalette (x,y,h: integer);
begin
  if bgBitmap.PixelFormat > pf8Bit then exit;
  SelectBitmap ('bg', bgBitmap);
  DrawPalette (tPalette(activePalette.palPalEntry), x,y,h);
  RestoreBitmap;
end;
//*********************************************************
procedure InitPalette;
var   ci: integer;
begin
  activePalette.palVersion := $300;
  activePalette.palNumEntries := palColors;

  SetPaletteColor (iBlack,     clBlack);
  SetPaletteColor (iBlue,      clBlue);
  SetPaletteColor (iGreen,     clGreen);
  SetPaletteColor (iCyan,      clCyan);
  SetPaletteColor (iRed,       clRed);
  SetPaletteColor (iMagenta,   clMagenta);
  SetPaletteColor (iBrown,     clBrown);
  SetPaletteColor (iLightGray, clLightGray);

  SetPaletteColor (iDarkGray,     clDarkGray);
  SetPaletteColor (iLightBlue,    clLightBlue);
  SetPaletteColor (iLightGreen,   clLightGreen);
  SetPaletteColor (iLightCyan,    clLightCyan);
  SetPaletteColor (iLightRed,     clLightRed);
  SetPaletteColor (iLightMagenta, clLightMagenta);
  SetPaletteColor (iYellow,       clYellow);
  SetPaletteColor (iWhite,        clWhite);

  for ci := 16 to palColors-1
  do with activePalette.palPalEntry[ci]
  do begin
    peRed   := ci;
    peGreen := ci;
    peBlue  := ci;
  end;
end;

//*********************************************************
// set given palette as active palette
//---------------------------------------------------------
procedure SetAllPalette (var pal: tPalette);
begin
{$ifdef test1}
  writeln (tst, 'Palette.SetAllPalette: ');
{$endif}

  Move (pal, activePalette.palPalEntry[0], palSize);
  ActivatePalette;
end;
//*********************************************************
// get active palette
//---------------------------------------------------------
procedure GetAllPalette (var pal: tPalette);
begin
  Move (activePalette.palPalEntry[0], pal, palSize);
end;
//*********************************************************
// set RGB color value of active color palette
//---------------------------------------------------------
procedure SetPaletteRGB (colorIndex: byte; r,g,b: byte);
begin
  with activePalette.palPalEntry[colorIndex]
  do begin
    peRed   := r;
    peGreen := g;
    peBlue  := b;
    peFlags := 0;
  end;
end;
//---------------------------------------------------------
procedure SetPaletteColor (colorIndex: byte; color: tColor);

begin
  with activePalette.palPalEntry[colorIndex], tColorRGBI(color)
  do begin
    peRed   := rb;
    peGreen := gb;
    peBlue  := bb;
    peFlags := 0;
  end;
end;
//---------------------------------------------------------
function GetPaletteColor (colorIndex: byte): tColor;
begin
  with activePalette.palPalEntry[colorIndex]
  do result := RGB (peRed, peGreen, peBlue);
end;

//=== bitmap handling =====================================

//*********************************************************
// Set given palette to given bitmap
// active palette will NOT be changed !
//---------------------------------------------------------
procedure SetBitmapPalette (bm: tBitmap; var pal: tPalette);
var   pv: tMaxLogPalette;
      ph: HPALETTE;
begin
{$ifdef test1}
  writeln (tst, 'Palette.SetBitmapPalette: ', HexString(bm));
{$endif}

  pv.palVersion := $300;
  pv.palNumEntries := 256;
  Move (pal, pv.palPalEntry[0], palSize);
  ph := Windows.CreatePalette(pLogPalette(@pv)^);
  if ph <> 0
  then bm.Palette := ph;
end;
//*********************************************************
// copy palette colors of given bitmap to given palette
//---------------------------------------------------------
procedure GetBitmapPalette (bm: tBitmap; var pal: tPalette);
begin
  FillChar (pal, sizeof (pal), 0);
  if bm.palette > 0 then
  GetPaletteEntries (bm.Palette, 0, 256, pal);
end;
//*********************************************************
// set palette of a given bitmap to other bitmap
//---------------------------------------------------------
procedure CopyBitmapPalette (bm, dm: tBitmap);
var   helpPal: tPalette;
begin
  if bm.palette = 0 then exit;       // no palette ?
  if dm.palette = 0 then exit;       // no palette ?
  GetBitmapPalette (bm, helpPal);
  SetBitmapPalette (dm, helpPal);
end;
//*********************************************************
// use palette of a given bitmap
// (copy to bgBitmap and activate)
//---------------------------------------------------------
procedure UseBitmapPalette (bm: tBitmap);
begin
{$ifdef test1}
  writeln (tst, 'Palette.UseBitmapPalette: ', HexString(bm));
{$endif}

  if bm.palette = 0 then exit;       // no palette ?
  // copy bitmap palette entries to activePalette
//  with activePalette
//  do GetPaletteEntries(bm.palette, 0, palNumEntries, palPalEntry);
  GetBitmapPalette (bm, tPalette(activePalette.palPalEntry));

  ActivatePalette;           // activate color palette
end;

//=== file handling =======================================

//*********************************************************
// load palette from file
//---------------------------------------------------------
procedure LoadPalette (var pal: tPalette; filename: string);
var   palFile: File;
begin
  AssignFile (palFile, filename);
  Reset      (palFile, 1);
  BlockRead  (palFile, pal, SizeOf(pal));
  CloseFile  (palFile);
end;
//*********************************************************
// save palette to file
//---------------------------------------------------------
procedure SavePalette (var pal: tPalette; filename: string);
var   palFile: File;
begin
  AssignFile (palFile, filename);
  Rewrite    (palFile, 1);
  BlockWrite (palFile, pal, SizeOf(pal));
  CloseFile  (palFile);
end;

//*********************************************************
// set system palette as active palette
//---------------------------------------------------------
procedure GetSystemPalette;
begin
  activePalette.palVersion := $300;
  activePalette.palNumEntries := 256;
  if GetSystemPaletteEntries (GetDC (GetActiveWindow)
                             ,0,256,activePalette.palPalEntry) < 256
  then InitPalette;
end;
//*********************************************************
begin
  GetSystemPalette;
end.

