
//*********************************************************
//
//  UNIT:     V a l u e s            12.06.2000, 08.06.2003
//
//  handle value range and interaction, coded by Gerd Platl
//
//---------------------------------------------------------
//
//  Init                constructor
//  Clear               clear value list
//  Reset               reset list values
//  AppendValue         append new value
//
//  SetValue            set value
//  IncValue            increment value
//  RandomValues        create random values
//  Animate             animate random values
//
//  FloatValue          return actual float value
//  IntValue            return actual integer value
//  ByteValue           return actual byte value
//  ValueCount          return number of values
//
//  WriteValues         write value list
//  KeyDown             handle key input
//  Paint               paint value list
//
//*********************************************************
{$I 'comp_opt.inc'}

unit Values;

{$ifopt D+}
  {$define xtest1    main functions}
  {$define xtest2    sub functions}
  {$define xtest3    view key value}
{$endif}

interface

uses  Classes;

{$I float.inc}

const maxValueIndex = 20;

type  tValueRecord = record
        minV, maxV, range, actValue, step: float;
        digit: byte;
        name: string[20];
      end;

  pValues = ^tValues;
  tValues = class (tObject)
  private
    maxIndex: word;
    actIndex: word;
    valueArray: array [1..maxValueIndex] of tValueRecord;

  public
    constructor Init;
    procedure Clear;
    procedure Reset;
    procedure AppendValue (infoText: string;
                           minValue, maxValue: float;
                           const start: float = -9E20;
                           const digits: byte = 0);

    procedure SetValue (i: byte; v: float);
    procedure IncValue (i: byte; v: float);
    procedure RandomValues;
    procedure Animate;

    function FloatValue (index: word): float;
    function IntValue   (index: word): integer;
    function ByteValue  (index: word): byte;
    function ValueCount: Word;

    procedure WriteValues (var out: text);
    function  KeyDown (var key: Word;
                       var shiftState: TShiftState): word;
    procedure Paint;
  private
    procedure ChangeValue (value: float);
  end;

//*********************************************************
implementation

{$define graph}

uses  SysUtils, Windows,
      {$ifdef graph} Graph
      {$else}        ColorRGB, Gr32, GrWindow   {$endif}
      ,SysUtil2;

const steps: float = 40;

//---------------------------------------------------------
constructor tValues.Init;
begin
{$ifdef test1}
  writeln (tst, HexString(self)
    ,' Values.Init: ');
{$endif}
  Clear;
end;

//---------------------------------------------------------
procedure tValues.Clear;
begin
{$ifdef test1}
  writeln (tst, HexString(self)
    ,' Values.Clear: ');
{$endif}
  maxIndex := 0;
  actIndex := 1;
end;

//---------------------------------------------------------
procedure tValues.Reset;
var   ni: integer;
begin
{$ifdef test1}
  writeln (tst, HexString(self)
    ,' Values.Reset:  maxIndex=',maxIndex);
{$endif}
  for ni := 1 to maxIndex
  do with valueArray[ni] do actValue := (minV + maxV) / 2.0;
end;

//*********************************************************
// append new value to value list
// input:  infotext   value name
//         minValue
//         maxValue   upper limit
//         --optional--
//           start    start value
//           digits   number of digits after dot
//                    (only for float value display)
//---------------------------------------------------------
procedure tValues.AppendValue (infoText: string;
                               minValue, maxValue: float;
                               const start: float;
                               const digits: byte);
begin
{$ifdef test2}
  writeln (tst, HexString(self)
    ,' Values.AppendValue (', infoText, minValue:10:digits
                                       , maxValue:10:digits,')');
{$endif}
  inc (maxIndex);
  if maxIndex > maxValueIndex
  then raise ERangeError.CreateFmt(
      '%d is not within the valid range of %d..%d',
      [maxIndex, 1, maxValueIndex]);

  with valueArray[maxIndex]
  do begin
    minV := minValue;
    maxV := maxValue;
    range := maxV - minV;
    if (start < minV) or (start > maxV)
    then actValue := (maxV + minV) / 2.0
    else actValue := start;
    if (range >= 10) and (range <= 60)
    then step := 1.0
    else step := range / steps;
    name := infoText;
    digit := digits;
  end;
end;
//*********************************************************
// return number of added values
//---------------------------------------------------------
function tValues.ValueCount: Word;
begin
  result := maxIndex;
end;

//---------------------------------------------------------
procedure tValues.SetValue (i: byte; v: float);
begin
  with valueArray[i] do begin
    actValue := v;
    if      actValue < minV then actValue := minV
    else if actValue > maxV then actValue := maxV;
  end;
end;
//---------------------------------------------------------
procedure tValues.IncValue (i: byte; v: float);
begin
  with valueArray[i] do begin
    actValue := actValue + v;
    if      actValue < minV then actValue := minV
    else if actValue > maxV then actValue := maxV;
  end;
end;
//---------------------------------------------------------
procedure tValues.RandomValues;
var   ni: integer;
begin
  for ni := 1 to maxIndex
  do with valueArray[ni] do actValue := minV + range * random;
end;
//---------------------------------------------------------
procedure tValues.Animate;
begin
  with valueArray[1 + random (maxIndex)]
  do if random (100) >= 50
  then begin
    if actValue + step < maxV then actValue := actValue + step
  end
  else if actValue - step > minV then actValue := actValue - step;
end;
//*********************************************************
// return actual values...
//---------------------------------------------------------
function tValues.FloatValue (index: word): float;
begin
  result := valueArray[index].actValue;
end;
//---------------------------------------------------------
function tValues.IntValue (index: word): integer;
begin
  result := trunc(valueArray[index].actValue);
end;
//---------------------------------------------------------
function tValues.ByteValue (index: word): byte;
begin
  result := byte (trunc(valueArray[index].actValue));
end;

//*********************************************************
//  write value list to given to text output device
//  example:   valuePtr^.WriteValues (tst);
//---------------------------------------------------------
procedure tValues.WriteValues (var out: text);
var   ni: integer;
begin
  writeln (out, '-----Values-----');
  writeln (out, '  actIndex=',actIndex);
  writeln (out, ' i_______min_______max_____range_____value__info');
  for ni := 1 to maxIndex
  do with valueArray[ni] do writeln (out, ni:2
  , minV:10:digit, maxV:10:2, range:10:2, actValue:10:2, '  ',name);
  writeln (out, '----------------');
  flush (out);
end;

//*********************************************************
procedure tValues.ChangeValue (value: float);
begin
  with valueArray[actIndex]
  do if   value < minV then actValue := minV
  else if value > maxV then actValue := maxV
  else                      actValue := value;
end;
//*********************************************************
// handle key input
// input:  key   virtual key, see TWinControl.OnKeyDown
// return: word  =0: if value was not changed
//               =index: index of changed value
//---------------------------------------------------------
function tValues.KeyDown (var key: Word;
                          var shiftState: TShiftState): word;
begin
{$ifdef test3}
  writeln (tst, 'Values.KeyDown: ',key:3);
{$endif}
  Result := 0;
  if maxIndex <= 0 then exit;
  if shiftState = [] then
  case key of
    vk_UP    : actIndex := (actIndex+maxIndex-2) mod maxIndex + 1;
    vk_DOWN  : actIndex := (actIndex mod maxIndex) + 1;
    vk_HOME  : actIndex := 1;
    vk_END   : actIndex := maxIndex;
    else with valueArray[actIndex]
    do case key of
      vk_SUBTRACT,
      vk_LEFT  : ChangeValue (actValue - step);
      vk_ADD,
      vk_RIGHT : ChangeValue (actValue + step);
      vk_NEXT  : actValue := minV;
      vk_PRIOR : actValue := maxV;
      vk_MULTIPLY,
      vk_CLEAR : ChangeValue ((minV + maxV) / 2);
      else Result := actIndex;
    end;
  end;
  if shiftState = [ssShift] then
  with valueArray[actIndex]
  do case key of
    vk_LEFT  : ChangeValue (actValue - 1);
    vk_RIGHT : ChangeValue (actValue + 1);
  end;
end;

{$ifdef graph}
//*********************************************************
//  draw all actual values and name of activ value
//---------------------------------------------------------
procedure tValues.Paint;
var   ni, x1,x2, xi: integer;   valStr: string;
begin
  if maxIndex = 0 then exit;           // nothing to paint ?
{$ifdef test2}
  writeln (tst, 'Values.Paint:  maxIndex=',maxIndex);
{$endif}

  x1 := GetMaxX div 2;
  x2 := GetMaxX - 10;
  for ni := 1 to maxIndex
  do with valueArray[ni]
  do begin
    if actIndex = ni then SetColor(clYellow) else SetColor (clOrange);
    xi := x1 + round((actValue - minV) * (x2-x1) / range);
    DrawHLine (x1, xi, 14 + (ni+1) * 4);
  end;
  with valueArray[actIndex]
  do begin
    if digit = 0
    then valStr := FloatToStrF(trunc(actValue), ffNumber, 5, digit)
    else valStr := FloatToStrF(actValue, ffNumber, 5, digit);
    ShadowText (x1, maxIndex * 4 + 22, name + ' ' + valStr);
  end;
end;
{$else}
//*********************************************************
//  draw all actual values and name of activ value
//---------------------------------------------------------
procedure tValues.Paint;
var   ni, x1,x2, xi, yp: integer;   valStr: string;
      color: tColor32;
begin
  if maxIndex = 0 then exit;           // nothing to paint ?
{$ifdef test2}
  writeln (tst, 'Values.Paint:  maxIndex=',maxIndex);
{$endif}
  try
  x1 := bm32.Width div 2;
  x2 := bm32.Height - 10;
  for ni := 1 to maxIndex
  do with valueArray[ni]
  do begin
    if actIndex = ni then color := Yellow else color := Orange;
    xi := x1 + round((actValue - minV) * (x2-x1) / range);
    yp := 14 + (ni+1) * 4;
    bm32.LineAS (x1, yp, xi, yp, color);
  end;
  with valueArray[actIndex]
  do begin
    if digit = 0
    then valStr := FloatToStrF(trunc(actValue), ffNumber, 5, digit)
    else valStr := FloatToStrF(actValue, ffNumber, 5, digit);
    grWin.SetTextColor (yellow);
    grWin.ShowText (x1, maxIndex * 4 + 22, name + ' ' + valStr);
  end;
  except
  end;
end;
{$endif}

end.
